#!/usr/bin/python
# -*- encoding: utf-8 -*-
"""

"""
__author__ = 'Martin Martimeo <martin@martimeo.de>'
__date__ = '29.04.13 - 15:37'

from .api import ApiManager

__all__ = ['ApiManager']
__version__ = '0.4.4'



#!/usr/bin/python
# -*- encoding: utf-8 -*-
"""
    Tornado Restless BaseHandler

    Handles all registered blueprints, you may override this class and
     use the modification via create_api_blueprint(handler_class=...)
"""
import inspect
from json import loads
import logging
from math import ceil
from traceback import print_exception
from urllib.parse import parse_qs
import sys
import itertools

from sqlalchemy import inspect as sqinspect
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound, UnmappedInstanceError, MultipleResultsFound
from sqlalchemy.util import memoized_instancemethod, memoized_property
from tornado.web import RequestHandler, HTTPError

from .convert import to_dict, to_filter
from .errors import IllegalArgumentError, MethodNotAllowedError, ProcessingException
from .wrapper import SessionedModelWrapper


__author__ = 'Martin Martimeo <martin@martimeo.de>'
__date__ = '26.04.13 - 22:09'


class BaseHandler(RequestHandler):
    """
        Basic Blueprint for a sqlalchemy model

        Subclass of :class:`tornado.web.RequestHandler` that handles web requests.

        Overwrite :func:`get() <get>` / :func:`post() <post>` / :func:`put() <put>` /
        :func:`patch() <patch>` / :func:`delete() <delete>` if you want complete customize handling of the methods.
        Note that the default implementation of this function check for the allowness and then call depending on
        the instance_id parameter the associated _single / _many method, so you probably want to call super()

        If you just want to customize the handling of the methods overwrite method_single or method_many.

        If you want completly disable a method overwrite the SUPPORTED_METHODS constant
    """

    ID_SEPARATOR = ","
    SUPPORTED_METHODS = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']

    # noinspection PyMethodOverriding
    def initialize(self,
                   model,
                   manager,
                   methods: set,
                   preprocessor: dict,
                   postprocessor: dict,
                   allow_patch_many: bool,
                   allow_method_override: bool,
                   validation_exceptions,
                   exclude_queries: bool,
                   exclude_hybrids: bool,
                   include_columns: list,
                   exclude_columns: list,
                   results_per_page: int,
                   max_results_per_page: int):
        """

        Init of the handler, derives arguments from api create_api_blueprint

        :param model: The sqlalchemy model
        :param manager: The tornado_restless Api Manager
        :param methods: Allowed methods for this model
        :param preprocessor: A dictionary of preprocessor functions
        :param postprocessor: A dictionary of postprocessor functions
        :param allow_patch_many: Allow PATCH with multiple datasets
        :param allow_method_override: Support X-HTTP-Method-Override Header
        :param validation_exceptions:
        :param exclude_queries: Don't execude dynamic queries (like from associations or lazy relations)
        :param exclude_hybrids: When exclude_queries is True and exclude_hybrids is False, hybrids are still included.
        :param include_columns: Whitelist of columns to be included
        :param exclude_columns: Blacklist of columns to be excluded
        :param results_per_page: The default value of how many results are returned per request
        :param max_results_per_page: The hard upper limit of resutest per page

        :reqheader X-HTTP-Method-Override: If allow_method_override is True, this header overwrites the request method
        """

        # Override Method if Header provided
        if allow_method_override and 'X-HTTP-Method-Override' in self.request.headers:
            self.request.method = self.request.headers['X-HTTP-Method-Override']

        super(BaseHandler, self).initialize()

        self.model = SessionedModelWrapper(model, manager.session_maker())
        self.pk_length = len(sqinspect(model).primary_key)
        self.methods = [method.lower() for method in methods]
        self.allow_patch_many = allow_patch_many
        self.validation_exceptions = validation_exceptions

        self.preprocessor = preprocessor
        self.postprocessor = postprocessor

        self.results_per_page = results_per_page
        self.max_results_per_page = max_results_per_page

        self.include = self.parse_columns(include_columns)
        self.exclude = self.parse_columns(exclude_columns)

        self.to_dict_options = {'execute_queries': not exclude_queries, 'execute_hybrids': not exclude_hybrids}

    def prepare(self):
        """
            Prepare the request
        """
        self._call_preprocessor()

    def on_finish(self):
        """
            Finish the request
        """
        self._call_postprocessor()

    def parse_columns(self, strings: list) -> dict:
        """
            Parse a list of column names (name1, name2, relation.name1, ...)

            :param strings: List of Column Names
            :return:
        """
        columns = {}

        # Strings
        if strings is None:
            return None

        # Parse
        for column in [column.split(".", 1) for column in strings]:
            if len(column) == 1:
                columns[column[0]] = True
            else:
                columns.setdefault(column[0], []).append(column[1])

        # Now parse relations
        for (key, item) in columns.items():
            if isinstance(item, list):
                columns[key] = itertools.chain.from_iterable(self.parse_columns(strings) for strings in item)

        # Return
        return columns

    def get_filters(self):
        """
            Returns a list of filters made by the query argument

            :query filters: list of filters
            :query order_by: list of orderings
        """

        # Get all provided filters
        argument_filters = self.get_query_argument("filters", [])

        # Get all provided orders
        argument_orders = self.get_query_argument("order_by", [])

        return to_filter(self.model.model, argument_filters, argument_orders)

    def write_error(self, status_code: int, **kwargs):
        """
            Encodes any exceptions thrown to json

            SQLAlchemyError will be encoded as 400 / SQLAlchemy: Bad Request
            Errors from the restless api as 400 / Restless: Bad Arguments
            ProcessingException will be encoded with status code / ProcessingException: Stopped Processing
            Any other exceptions will occur as an 500 exception

            :param status_code: The Status Code in Response
            :param kwargs: Additional Parameters
        """
        if 'exc_info' in kwargs:
            exc_type, exc_value = kwargs['exc_info'][:2]
            if status_code >= 300:
                print_exception(*kwargs['exc_info'])
            if issubclass(exc_type, UnmappedInstanceError):
                self.set_status(400, reason='SQLAlchemy: Unmapped Instance')
                self.finish(dict(type=exc_type.__module__ + "." + exc_type.__name__,
                                 message="%s" % exc_value))
            elif issubclass(exc_type, SQLAlchemyError):
                if issubclass(exc_type, NoResultFound):
                    status = 404
                    reason = message = 'No result found'
                elif issubclass(exc_type, MultipleResultsFound):
                    status = 400
                    reason = 'SQLAlchemy: Bad Request'
                    message = 'Multiple results found'
                else:
                    status = 400
                    reason = 'SQLAlchemy: Bad Request'
                    message = "%s" % exc_value
                self.set_status(status, reason=reason)
                self.finish(dict(type=exc_type.__module__ + "." + exc_type.__name__,
                                 message=message))
            elif issubclass(exc_type, IllegalArgumentError):
                self.set_status(400, reason='Restless: Bad Arguments')
                self.finish(dict(type=exc_type.__module__ + "." + exc_type.__name__,
                                 message="%s" % exc_value))
            elif issubclass(exc_type, ProcessingException):
                self.set_status(status_code,
                                reason='ProcessingException: %s' % (exc_value.reason or "Stopped Processing"))
                self.finish(dict(type=exc_type.__module__ + "." + exc_type.__name__,
                                 message="%s" % exc_value))
            elif issubclass(exc_type, HTTPError) and exc_value.reason:
                self.set_status(status_code, reason=exc_value.reason)
                self.finish(dict(type=exc_type.__module__ + "." + exc_type.__name__,
                                 message="%s" % exc_value, **exc_value.__dict__))
            else:
                super().write_error(status_code, **kwargs)
        else:
            super().write_error(status_code, **kwargs)

    def patch(self, instance_id: str=None):
        """
            PATCH (update instance) request

            :param instance_id: query argument of request
            :type instance_id: comma seperated string list

            :statuscode 403: PATCH MANY disallowed
            :statuscode 405: PATCH disallowed
        """

        if not 'patch' in self.methods:
            raise MethodNotAllowedError(self.request.method)

        self._call_preprocessor(search_params=self.search_params)

        if instance_id is None:
            if self.allow_patch_many:
                result = self.patch_many()
            else:
                raise MethodNotAllowedError(self.request.method, status_code=403)
        else:
            result = self.patch_single(self.parse_pk(instance_id))

        self._call_postprocessor(result=result)
        self.finish(result)

    def patch_many(self) -> dict:
        """
            Patch many instances

            :statuscode 201: instances successfull modified

            :query limit: limit the count of modified instances
            :query single: If true sqlalchemy will raise an error if zero or more than one instances would be modified
        """

        # Flush
        self.model.session.flush()

        # Get values
        values = self.get_argument_values()

        # Filters
        filters = self.get_filters()

        # Limit
        limit = self.get_query_argument("limit", None)

        # Call Preprocessor
        self._call_preprocessor(filters=filters, data=values)

        # Modify Instances
        if self.get_query_argument("single", False):
            instances = [self.model.one(filters=filters)]
            for instance in instances:
                for (key, value) in values.items():
                    logging.debug("%s => %s" % (key, value))
                    setattr(instance, key, value)
            num = 1
        else:
            num = self.model.update(values, limit=limit, filters=filters)

        # Commit
        self.model.session.commit()

        # Result
        self.set_status(201, "Patched")
        return {'num_modified': num}

    def patch_single(self, instance_id: list) -> dict:
        """
            Patch one instance

            :param instance_id: query argument of request
            :type instance_id: list of primary keys

            :statuscode 201: instance successfull modified
            :statuscode 404: Error
        """
        try:
            with self.model.session.begin_nested():
                values = self.get_argument_values()

                # Call Preprocessor
                self._call_preprocessor(instance_id=instance_id, data=values)

                # Get Instance
                instance = self.model.get(*instance_id)

                # Set Values
                for (key, value) in values.items():
                    self.logger.debug("%r.%s => %s" % (instance, key, value))
                    setattr(instance, key, value)

                # Flush
                try:
                    self.model.session.flush()
                except SQLAlchemyError as ex:
                    logging.exception(ex)
                    self.model.session.rollback()
                    self.send_error(status_code=400, exc_info=sys.exc_info())
                    return

                # Refresh
                self.model.session.refresh(instance)

                # Set Status
                self.set_status(201, "Patched")

                # To Dict
                return self.to_dict(instance)
        except SQLAlchemyError as ex:
            logging.exception(ex)
            self.send_error(status_code=400, exc_info=sys.exc_info())
        finally:
            # Commit
            self.model.session.commit()

    def delete(self, instance_id: str=None):
        """
            DELETE (delete instance) request

            :param instance_id: query argument of request
            :type instance_id: comma seperated string list

            :statuscode 403: DELETE MANY disallowed
            :statuscode 405: DELETE disallowed
        """

        if not 'delete' in self.methods:
            raise MethodNotAllowedError(self.request.method)

        # Call Preprocessor
        self._call_preprocessor(search_params=self.search_params)

        if instance_id is None:
            if self.allow_patch_many:
                result = self.delete_many()
            else:
                raise MethodNotAllowedError(self.request.method, status_code=403)
        else:
            result = self.delete_single(self.parse_pk(instance_id))

        self._call_postprocessor(result=result)
        self.finish(result)

    def delete_many(self) -> dict:
        """
            Remove many instances

            :statuscode 200: instances successfull removed

            :query limit: limit the count of deleted instances
            :query single: If true sqlalchemy will raise an error if zero or more than one instances would be deleted
        """

        # Flush
        self.model.session.flush()

        # Filters
        filters = self.get_filters()

        # Limit
        limit = self.get_query_argument("limit", None)

        # Call Preprocessor
        self._call_preprocessor(filters=filters)

        # Modify Instances
        if self.get_query_argument("single", False):
            instance = self.model.one(filters=filters)
            self.model.session.delete(instance)
            self.model.session.commit()
            num = 1
        else:
            num = self.model.delete(limit=limit, filters=filters)

        # Commit
        self.model.session.commit()

        # Result
        self.set_status(200, "Removed")
        return {'num_removed': num}

    def delete_single(self, instance_id: list) -> dict:
        """
            Get one instance

            :param instance_id: query argument of request
            :type instance_id: list of primary keys

            :statuscode 204: instance successfull removed
        """

        # Call Preprocessor
        self._call_preprocessor(instance_id=instance_id)

        # Get Instance
        instance = self.model.get(*instance_id)

        # Trigger deletion
        self.model.session.delete(instance)
        self.model.session.commit()

        # Status
        self.set_status(204, "Instance removed")
        return {}

    def put(self, instance_id: str=None):
        """
            PUT (update instance) request

            :param instance_id: query argument of request
            :type instance_id: comma seperated string list

            :statuscode 403: PUT MANY disallowed
            :statuscode 404: Error
            :statuscode 405: PUT disallowed
        """

        if not 'put' in self.methods:
            raise MethodNotAllowedError(self.request.method)

        # Call Preprocessor
        self._call_preprocessor(search_params=self.search_params)

        if instance_id is None:
            if self.allow_patch_many:
                result = self.put_many()
            else:
                raise MethodNotAllowedError(self.request.method, status_code=403)
        else:
            result = self.put_single(self.parse_pk(instance_id))

        self._call_postprocessor(result=result)
        self.finish(result)

    put_many = patch_many
    put_single = patch_single

    def post(self, instance_id: str=None):
        """
            POST (new input) request

            :param instance_id: (ignored)

            :statuscode 204: instance successfull created
            :statuscode 404: Error
            :statuscode 405: POST disallowed
        """

        if not 'post' in self.methods:
            raise MethodNotAllowedError(self.request.method)

        # Call Preprocessor
        self._call_preprocessor(search_params=self.search_params)

        result = self.post_single()

        self._call_postprocessor(result=result)
        self.finish(result)

    def post_single(self):
        """
            Post one instance
        """

        try:
            values = self.get_argument_values()

            # Call Preprocessor
            self._call_preprocessor(data=values)

            # Create Instance
            instance = self.model(**values)

            # Flush
            self.model.session.commit()

            # Refresh
            self.model.session.refresh(instance)

            # Set Status
            self.set_status(201, "Created")

            # To Dict
            return self.to_dict(instance)
        except SQLAlchemyError:
            self.send_error(status_code=400, exc_info=sys.exc_info())
            self.model.session.rollback()
        finally:
            # Commit
            self.model.session.commit()

    @memoized_instancemethod
    def get_content_encoding(self) -> str:
        """
        Get the encoding the client sends us for encoding request.body correctly

        :reqheader Content-Type: Provide a charset in addition for decoding arguments.
        """

        content_type_args = {k.strip(): v for k, v in parse_qs(self.request.headers['Content-Type']).items()}
        if 'charset' in content_type_args and content_type_args['charset']:
            return content_type_args['charset'][0]
        else:
            return 'latin1'

    @memoized_instancemethod
    def get_body_arguments(self) -> dict:
        """
            Get arguments encode as json body

            :statuscode 415: Content-Type mismatch

            :reqheader Content-Type: application/x-www-form-urlencoded or application/json
        """

        self.logger.debug(self.request.body)

        content_type = self.request.headers.get('Content-Type')
        if 'www-form-urlencoded' in content_type:
            payload = self.request.arguments
            for key, value in payload.items():
                if len(value) == 0:
                    payload[key] = None
                elif len(value) == 1:
                    payload[key] = str(value[0], encoding=self.get_content_encoding())
                else:
                    payload[key] = [str(value, encoding=self.get_content_encoding()) for value in value]
            return payload
        elif 'application/json' in content_type:
            return loads(str(self.request.body, encoding=self.get_content_encoding()))
        else:
            raise HTTPError(415, content_type=content_type)

    def get_body_argument(self, name: str, default=RequestHandler._ARG_DEFAULT):
        """
        Get an argument named key from json encoded body

        :param name: Name of argument
        :param default: Default value, if not provided HTTPError 404 is raised
        :return:

        :statuscode 404: Missing Argument
        """
        arguments = self.get_body_arguments()
        if name in arguments:
            return arguments[name]
        elif default is RequestHandler._ARG_DEFAULT:
            raise HTTPError(400, "Missing argument %s" % name)
        else:
            return default

    @property
    def search_params(self) -> dict:
        """
            The 'q' Dictionary
        """
        try:
            return self._search_params
        except AttributeError:
            self._search_params = loads(self.get_argument("q", default="{}"))
            return self._search_params

    def get_query_argument(self, name: str, default=RequestHandler._ARG_DEFAULT):
        """
        Get an argument named key from json encoded body

        :param name:
        :param default:
        :return:
        :raise: 400 Missing Argument

        :query q: The query argument
        """

        if name in self.search_params:
            return self.search_params[name]
        elif default is RequestHandler._ARG_DEFAULT:
            raise HTTPError(400, "Missing argument %s" % name)
        else:
            return default

    def get_argument(self, name: str, *args, **kwargs):
        """
            On PUT/PATCH many request parameter may be located in body instead of query

            :param name: Name of argument
            :param args: Additional position arguments @see tornado.web.RequestHandler.get_argument
            :param kwargs: Additional keyword arguments @see tornado.web.RequestHandler.get_argument
        """
        try:
            return super().get_argument(name, *args, **kwargs)
        except HTTPError:
            if name == "q" and self.request.method in ['PUT', 'PATCH']:
                return self.get_body_argument(name, *args, **kwargs)
            else:
                raise

    def get_argument_values(self):
        """
            Get all values provided via arguments

            :query q: (ignored)
        """

        # Include Columns
        if self.include is not None:
            values = {k: self.get_body_argument(k) for k in self.include}
        else:
            values = {k: v for k, v in self.get_body_arguments().items()}

        # Exclude "q"
        if "q" in values:
            del values["q"]

        # Exclude Columns
        if self.exclude is not None:
            for column in list(self.exclude):
                if column in values:
                    del values[column]

        # Silently Ignore proxies
        for proxy in self.model.proxies:
            if proxy.key in values:
                self.logger.debug("Skipping proxy: %s" % proxy.key)
                del values[proxy.key]

        # Silently Ignore hybrids
        for hybrid in self.model.hybrids:
            if hybrid.key in values:
                self.logger.debug("Skipping hybrid: %s" % hybrid.key)
                del values[hybrid.key]

        # Handle Relations extra
        values_relations = {}
        for relation_key, relation in self.model.relations.items():
            if relation_key in values:
                values_relations[relation_key] = values[relation_key]
                del values[relation_key]

        # Check Columns
        #for column in values:
        #    if not column in self.model.column_names:
        #        raise IllegalArgumentError("Column '%s' not defined for model %s" % (column, self.model.model))

        return values

    def get(self, instance_id: str=None):
        """
            GET request

            :param instance_id: query argument of request
            :type instance_id: comma seperated string list

            :statuscode 405: GET disallowed
        """

        if not 'get' in self.methods:
            raise MethodNotAllowedError(self.request.method)

        # Call Preprocessor
        self._call_preprocessor(search_params=self.search_params)

        if instance_id is None:
            result = self.get_many()
        else:
            result = self.get_single(self.parse_pk(instance_id))

        self._call_postprocessor(result=result)
        self.finish(result)

    def get_single(self, instance_id: list) -> dict:
        """
            Get one instance

            :param instance_id: query argument of request
            :type instance_id: list of primary keys
        """

        # Call Preprocessor
        self._call_preprocessor(instance_id=instance_id)

        # Get Instance
        instance = self.model.get(*instance_id)

        # To Dict
        return self.to_dict(instance)

    def get_many(self) -> dict:
        """
            Get all instances

            Note that it is possible to provide offset and page as argument then
            it will return instances of the nth page and skip offset items

            :statuscode 400: if results_per_page > max_results_per_page or offset < 0

            :query results_per_page: Overwrite the returned results_per_page
            :query offset: Skip offset instances
            :query page: Return nth page
            :query limit: limit the count of modified instances
            :query single: If true sqlalchemy will raise an error if zero or more than one instances would be deleted
        """

        # All search params
        search_params = {'single': self.get_query_argument("single", False),
                         'results_per_page': int(self.get_argument("results_per_page", self.results_per_page)),
                         'offset': int(self.get_query_argument("offset", 0))}

        # Results per Page Check
        if search_params['results_per_page'] > self.max_results_per_page:
            raise IllegalArgumentError("request.results_per_page > application.max_results_per_page")

        # Offset & Page
        page = int(self.get_argument("page", '1')) - 1
        search_params['offset'] += page * search_params['results_per_page']
        if search_params['offset'] < 0:
            raise IllegalArgumentError("request.offset < 0")

        # Limit
        search_params['limit'] = self.get_query_argument("limit", search_params['results_per_page'] or None)

        # Filters
        filters = self.get_filters()

        # Call Preprocessor
        self._call_preprocessor(filters=filters, search_params=search_params)

        # Num Results
        num_results = self.model.count(filters=filters)
        if search_params['results_per_page']:
            total_pages = ceil(num_results / search_params['results_per_page'])
        else:
            total_pages = 1

        # Get Instances
        if search_params['single']:
            instance = self.model.one(offset=search_params['offset'],
                                      filters=filters)
            return self.to_dict(instance)
        else:
            instances = self.model.all(offset=search_params['offset'],
                                       limit=search_params['limit'],
                                       filters=filters)
            return {'num_results': num_results,
                    "total_pages": total_pages,
                    "page": page + 1,
                    "objects": self.to_dict(instances)}

    def _call_preprocessor(self, *args, **kwargs):
        """
            Calls a preprocessor with args and kwargs
        """
        func_name = inspect.stack()[1][3]

        if func_name in self.preprocessor:
            for func in self.preprocessor[func_name]:
                func(*args, model=self.model, handler=self, **kwargs)

    def _call_postprocessor(self, *args, **kwargs):
        """
            Calls a postprocessor with args and kwargs
        """
        func_name = inspect.stack()[1][3]

        if func_name in self.postprocessor:
            for func in self.postprocessor[func_name]:
                func(*args, model=self.model, handler=self, **kwargs)

    @memoized_property
    def logger(self):
        """
            Tornado Restless Logger
        """
        return logging.getLogger('tornado.restless')

    def to_dict(self, instance):
        """
            Wrapper to convert.to_dict with arguments from blueprint init

            :param instance: Instance to be translated
        """
        return to_dict(instance,
                       include=self.include,
                       exclude=self.exclude,
                       options=self.to_dict_options)

    def parse_pk(self, instance_id):
        return instance_id.split(self.ID_SEPARATOR, self.pk_length - 1)



#!/usr/bin/python
# -*- encoding: utf-8 -*-
"""

"""
from collections import namedtuple
import inspect
import logging

from sqlalchemy import inspect as sqinspect
from sqlalchemy.exc import NoInspectionAvailable
from sqlalchemy.ext.associationproxy import AssociationProxy
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import ColumnProperty, Query
from sqlalchemy.orm.attributes import QueryableAttribute
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.interfaces import MapperProperty
from sqlalchemy.orm.properties import RelationshipProperty
from sqlalchemy.sql.operators import is_ordering_modifier
from sqlalchemy.util import memoized_property


__author__ = 'Martin Martimeo <martin@martimeo.de>'
__date__ = '27.04.13 - 00:14'


def _filter(instance, condition) -> dict:
    """
        Filter properties of instace based on condition

        :param instance:
        :param condition:
        :rtype: dict
    """

    # Use iterate_properties when available
    if hasattr(instance, 'iterate_properties'):
        return {field.key: field for field in instance.iterate_properties
                if condition(field)}

    # Try sqlalchemy inspection
    try:
        inspection = sqinspect(instance)
        if hasattr(inspection, "all_orm_descriptors"):
            return {field.key: field for key, field in inspection.all_orm_descriptors.items()
                    if condition(field)}
        elif hasattr(inspection, "attrs"):
            return {field.key: field for key, field in inspection.attrs.items()
                    if condition(field)}
        else:
            raise NoInspectionAvailable()
    # Use Inspect
    except NoInspectionAvailable:
        return {field.key: field for key, field in inspect.getmembers(instance)
                if condition(field)}


def _is_ordering_expression(expression):
    """
        Test an expression whether it is an ordering clause
    """

    if hasattr(expression, 'operator') and is_ordering_modifier(expression.operator):
        return True

    if hasattr(expression, 'modifier') and is_ordering_modifier(expression.modifier):
        return True

    return False


class ModelWrapper(object):
    """
        Wrapper around sqlalchemy model for having some easier functions
    """

    def __init__(self, model):
        self.model = model

    @property
    def __name__(self):
        return self.model.__name__

    @property
    def __tablename__(self):
        return self.model.__tablename__

    @property
    def __collectionname__(self):
        try:
            return self.model.__collectionname__
        except AttributeError:
            logging.warning("Missing collection name for %s using tablename" % self.model.__name__)
            return self.model.__tablename__

    @staticmethod
    def get_primary_keys(instance) -> dict:
        """
            Returns the primary keys

            Inspired by flask-restless.helpers.primary_key_names
        """
        return _filter(instance, lambda field: isinstance(field, ColumnProperty) and field.primary_key or (
            isinstance(field, QueryableAttribute) and isinstance(field.property, ColumnProperty) and
            hasattr(field.property.columns[0], 'primary_key') and field.property.columns[0].primary_key))

    @memoized_property
    def primary_keys(self):
        """
        @see get_primary_keys
        """
        return self.get_primary_keys(self.model)

    primary_keys.__doc__ = get_primary_keys.__func__.__doc__

    @staticmethod
    def get_unique_keys(instance) -> dict:
        """
            Returns the primary keys

            Inspired by flask-restless.helpers.primary_key_names
        """
        return _filter(instance, lambda field: isinstance(field, ColumnProperty) and field.unique or (
            isinstance(field, QueryableAttribute) and isinstance(field.property, ColumnProperty) and
            hasattr(field.property.columns[0], 'unique') and field.property.columns[0].unique))

    @memoized_property
    def unique_keys(self):
        """
        @see get_primary_keys
        """
        return self.get_unique_keys(self.model)

    unique_keys.__doc__ = get_unique_keys.__func__.__doc__

    @staticmethod
    def get_foreign_keys(instance) -> list:
        """
            Returns the foreign keys

            Inspired by flask-restless.helpers.primary_key_names
        """
        return {field.key: field for key, field in inspect.getmembers(instance)
                if isinstance(field, QueryableAttribute)
                   and isinstance(field.property, ColumnProperty)
        and field.foreign_keys}

    @memoized_property
    def foreign_keys(self):
        """
        @see get_foreign_keys
        """
        return self.get_foreign_keys(self.model)

    foreign_keys.__doc__ = get_foreign_keys.__func__.__doc__

    @staticmethod
    def get_columns(instance) -> dict:
        """
            Returns the columns objects of the model
        """
        return _filter(instance, lambda field: isinstance(field, ColumnProperty) or (
            isinstance(field, QueryableAttribute) and isinstance(field.property, ColumnProperty)))

    @memoized_property
    def columns(self):
        """
        @see get_columns
        """
        return self.get_columns(self.model)

    columns.__doc__ = get_columns.__func__.__doc__

    @staticmethod
    def get_attributes(instance) -> dict:
        """
            Returns the attributes of the model
        """
        return _filter(instance,
                       lambda field: isinstance(field, MapperProperty) or isinstance(field, QueryableAttribute))

    @memoized_property
    def attributes(self):
        """
        @see get_attributes
        """
        return self.get_attributes(self.model)

    attributes.__doc__ = get_attributes.__func__.__doc__

    @staticmethod
    def get_relations(instance) -> dict:
        """
            Returns the relations objects of the model
        """
        return _filter(instance, lambda field: isinstance(field, RelationshipProperty) or (
            isinstance(field, QueryableAttribute) and isinstance(field.property, RelationshipProperty)))

    @memoized_property
    def relations(self):
        """
        @see get_relations
        """
        return self.get_relations(self.model)

    relations.__doc__ = get_relations.__func__.__doc__

    @staticmethod
    def get_hybrids(instance) -> list:
        """
            Returns the relations objects of the model
        """
        Proxy = namedtuple('Proxy', ['key', 'field'])
        # Try sqlalchemy inspection
        try:
            return [Proxy(key, field) for key, field in sqinspect(instance).all_orm_descriptors.items()
                    if isinstance(field, hybrid_property)]
        # Use Inspect
        except NoInspectionAvailable:
            return [Proxy(key, field) for key, field in inspect.getmembers(instance)
                    if isinstance(field, hybrid_property)]

    @memoized_property
    def hybrids(self) -> list:
        """
        @see get_hybrids
        """
        return self.get_hybrids(self.model)

    hybrids.__doc__ = get_hybrids.__func__.__doc__

    @staticmethod
    def get_proxies(instance) -> list:
        """
            Returns the proxies objects of the model

            Inspired by https://groups.google.com/forum/?fromgroups=#!topic/sqlalchemy/aDi_M4iH7d0
        """
        Proxy = namedtuple('Proxy', ['key', 'field'])
        if hasattr(instance, 'iterate_properties'):
            return [Proxy(key, field) for key, field in sqinspect(instance).all_orm_descriptors.items()
                    if isinstance(field, AssociationProxy)]
        else:
            return [Proxy(key, field) for key, field in inspect.getmembers(instance)
                    if isinstance(field, AssociationProxy)]

    @memoized_property
    def proxies(self):
        """
        @see get_proxies
        """
        return self.get_proxies(self.model)

    proxies.__doc__ = get_proxies.__func__.__doc__


class SessionedModelWrapper(ModelWrapper):
    """
        Wrapper around sqlalchemy model for having some easier functions
    """

    def __init__(self, model, session):
        super().__init__(model)
        self.session = session

    @staticmethod
    def _apply_kwargs(instance: Query, **kwargs) -> Query:
        for expression in kwargs.pop('filters', []):
            if _is_ordering_expression(expression):
                instance = instance.order_by(expression)
            else:
                instance = instance.filter(expression)

        if 'offset' in kwargs:
            offset = kwargs.pop('offset')
            foffset = lambda instance: instance.offset(offset)
        else:
            foffset = lambda instance: instance

        if 'limit' in kwargs:
            limit = kwargs.pop('limit')
            flimit = lambda instance: instance.limit(limit)
        else:
            flimit = lambda instance: instance

        instance = instance.filter_by(**kwargs)
        instance = foffset(instance)
        instance = flimit(instance)
        return instance

    def one(self, filters: list=(), **kwargs) -> object:
        """
            Gets one instance of the model filtered by filters

            :param filters: Filters and OrderBy Clauses
            :param kwargs: Additional filters passed to filter_by
            :keyword offset: Offset for request
        """
        if isinstance(self, SessionedModelWrapper):
            instance = self.session.query(self.model)
        else:
            instance = self

        return SessionedModelWrapper._apply_kwargs(instance, filters=filters, **kwargs).one()

    def all(self, filters: list=(), **kwargs) -> list:
        """
            Gets all instances of the query instance

            :param filters: Filters and OrderBy Clauses
            :param kwargs: Additional filters passed to filter_by
            :keyword limit: Limit for request
            :keyword offset: Offset for request
        """
        if isinstance(self, SessionedModelWrapper):
            instance = self.session.query(self.model)
        else:
            instance = self

        return SessionedModelWrapper._apply_kwargs(instance, filters=filters, **kwargs).all()

    def update(self, values: dict, filters: list=(), **kwargs) -> int:
        """
            Updates all instances of the model filtered by filters

            :param values: Dictionary of values
            :param filters: Filters and OrderBy Clauses
            :param kwargs: Additional filters passed to filter_by
            :keyword limit: Limit for request
            :keyword offset: Offset for request
        """
        if isinstance(self, SessionedModelWrapper):
            instance = self.session.query(self.model)
        else:
            instance = self

        return SessionedModelWrapper._apply_kwargs(instance, filters=filters, **kwargs).update(values)

    def delete(self, filters: list=(), **kwargs) -> int:
        """
            Delete all instances of the model filtered by filters

            :param values: Dictionary of values
            :param filters: Filters and OrderBy Clauses
            :param kwargs: Additional filters passed to filter_by
            :keyword limit: Limit for request
            :keyword offset: Offset for request
        """
        if isinstance(self, SessionedModelWrapper):
            instance = self.session.query(self.model)
        else:
            instance = self

        return SessionedModelWrapper._apply_kwargs(instance, filters=filters, **kwargs).delete()

    def count(self, filters: list=(), **kwargs) -> int:
        """
            Gets the instance count

            :param filters: Filters and OrderBy Clauses
            :param kwargs: Additional filters passed to filter_by
        """
        if isinstance(self, SessionedModelWrapper):
            instance = self.session.query(self.model)
        else:
            instance = self

        return SessionedModelWrapper._apply_kwargs(instance, filters=filters, **kwargs).order_by(False).count()

    def get(self, *pargs) -> object:
        """
            Gets one instance of the model based on primary_keys

            :param pargs: ident
            :raise NoResultFound: If no element has been received
        """
        if isinstance(self, SessionedModelWrapper):
            instance = self.session.query(self.model)
        else:
            instance = self

        if not isinstance(pargs, tuple):
            rtn = instance.get(*pargs)
        else:
            rtn = instance.get(pargs)

        if not rtn:
            raise NoResultFound("No element recieved for %s(%s)" % (self.__collectionname__, pargs))

        return rtn

    def __call__(self, **kwargs):
        instance = self.model()
        for key, value in kwargs.items():
            setattr(instance, key, value)
        self.session.add(instance)
        return instance



#!/usr/bin/python
# -*- encoding: utf-8 -*-
"""

"""
from datetime import datetime, date, time
from decimal import Decimal
import collections
import itertools

from sqlalchemy.orm import object_mapper
from sqlalchemy.orm.exc import UnmappedInstanceError
from sqlalchemy.orm.query import Query

from .errors import IllegalArgumentError, DictConvertionError
from .wrapper import ModelWrapper


__author__ = 'Martin Martimeo <martin@martimeo.de>'
__date__ = '23.05.13 - 17:41'

__datetypes__ = (datetime, time, date)
__basetypes__ = (str, int, bool, float)
__clsztypes__ = (Decimal, )


def to_filter(instance,
              filters=None,
              order_by=None):
    """
        Returns a list of filters made by arguments

        :param instance:
        :param filters: List of filters in restless 3-tuple op string format
        :param order_by: List of orders to be appended aswell
    """

    # Get all provided filters
    argument_filters = filters and filters or []

    # Parse order by as filters
    argument_orders = order_by and order_by or []
    for argument_order in argument_orders:
        direction = argument_order['direction']
        if direction not in ["asc", "desc"]:
            raise IllegalArgumentError("Direction unknown")
        argument_filters.append({'name': argument_order['field'], 'op': direction,
                                 'nullsfirst': argument_order.get('nullsfirst', False),
                                 'nullslast': argument_order.get('nullslast', False)})

    # Create Alchemy Filters
    alchemy_filters = []
    for argument_filter in argument_filters:

        # Resolve right attribute
        if "field" in argument_filter.keys():
            right = getattr(instance, argument_filter["field"])
        elif "val" in argument_filter.keys():
            right = argument_filter["val"]
        elif "value" in argument_filter.keys():  # Because we hate abbr sometimes ...
            right = argument_filter["value"]
        else:
            right = None

        # Operator
        op = argument_filter["op"]

        # Resolve left attribute
        if "name" not in argument_filter:
            raise IllegalArgumentError("Missing fieldname attribute 'name'")

        if "__" in argument_filter["name"] or "." in argument_filter["name"]:
            relation, _, name = argument_filter["name"].replace("__", ".").partition(".")
            left = getattr(instance, relation)
            op = "has"
            argument_filter["name"] = name
            argument_filter["op"] = "eq"
            right = to_filter(instance=left.property.mapper.class_, filters=[argument_filter])
        elif argument_filter["name"] == "~":
            left = instance
            op = "attr_is"
        else:
            left = getattr(instance, argument_filter["name"])

        # Operators from flask-restless
        if op in ["is_null"]:
            alchemy_filters.append(left.is_(None))
        elif op in ["is_not_null"]:
            alchemy_filters.append(left.isnot(None))
        elif op in ["is"]:
            alchemy_filters.append(left.is_(right))
        elif op in ["is_not"]:
            alchemy_filters.append(left.isnot(right))
        elif op in ["==", "eq", "equals", "equals_to"]:
            alchemy_filters.append(left == right)
        elif op in ["!=", "ne", "neq", "not_equal_to", "does_not_equal"]:
            alchemy_filters.append(left != right)
        elif op in [">", "gt"]:
            alchemy_filters.append(left > right)
        elif op in ["<", "lt"]:
            alchemy_filters.append(left < right)
        elif op in [">=", "ge", "gte", "geq"]:
            alchemy_filters.append(left >= right)
        elif op in ["<=", "le", "lte", "leq"]:
            alchemy_filters.append(left <= right)
        elif op in ["ilike"]:
            alchemy_filters.append(left.ilike(right))
        elif op in ["not_ilike"]:
            alchemy_filters.append(left.notilike(right))
        elif op in ["like"]:
            alchemy_filters.append(left.like(right))
        elif op in ["not_like"]:
            alchemy_filters.append(left.notlike(right))
        elif op in ["match"]:
            alchemy_filters.append(left.match(right))
        elif op in ["in"]:
            alchemy_filters.append(left.in_(right))
        elif op in ["not_in"]:
            alchemy_filters.append(left.notin_(right))
        elif op in ["has"] and isinstance(right, list):
            alchemy_filters.append(left.any(*right))
        elif op in ["has"]:
            alchemy_filters.append(left.has(right))
        elif op in ["any"]:
            alchemy_filters.append(left.any(right))

        # Additional Operators
        elif op in ["between"]:
            alchemy_filters.append(left.between(*right))
        elif op in ["contains"]:
            alchemy_filters.append(left.contains(right))
        elif op in ["startswith"]:
            alchemy_filters.append(left.startswith(right))
        elif op in ["endswith"]:
            alchemy_filters.append(left.endswith(right))

        # Order By Operators
        elif op in ["asc", "desc"]:
            if argument_filter.get("nullsfirst", False):
                alchemy_filters.append(getattr(left, op)().nullsfirst())
            elif argument_filter.get("nullslast", False):
                alchemy_filters.append(getattr(left, op)().nullslast())
            else:
                alchemy_filters.append(getattr(left, op)())

        # Additional Checks
        elif op in ["attr_is"]:
            alchemy_filters.append(getattr(left, right))
        elif op in ["method_is"]:
            alchemy_filters.append(getattr(left, right)())

        # Test comparator
        elif hasattr(left.comparator, op):
            alchemy_filters.append(getattr(left.comparator, op)(right))

        # Raise Exception
        else:
            raise IllegalArgumentError("Unknown operator")
    return alchemy_filters


def to_deep(include,
            exclude,
            key):
    """
        Extract the include/exclude information for key

        :param include: Columns and Relations that should be included for an instance
        :param exclude: Columns and Relations that should not be included for an instance
        :param key: The key that should be extracted
    """
    rtn = {}

    try:
        rtn['include'] = include.setdefault(key, False)
    except AttributeError:
        rtn['include'] = False

    try:
        rtn['exclude'] = exclude[key]
    except (TypeError, KeyError):
        rtn['exclude'] = None

    return rtn


def to_dict(instance,
            options=collections.defaultdict(bool),
            include=None,
            exclude=None):
    """
        Translates sqlalchemy instance to dictionary

        Inspired by flask-restless.helpers.to_dict

        :param instance:
        :param options: Dictionary of flags
                          * execute_queries: Execute Query Objects
                          * execute_hybrids: Execute Hybrids
        :param include: Columns and Relations that should be included for an instance
        :param exclude: Columns and Relations that should not be included for an instance
    """
    if exclude is not None and include is not None:
        raise ValueError('Cannot specify both include and exclude.')

    # None
    if instance is None:
        return None

    # Int / Float / Str
    if isinstance(instance, __basetypes__):
        return instance

    # Date & Time
    if isinstance(instance, __datetypes__):
        return instance.isoformat()

    # Any Dictionary
    if isinstance(instance, dict) or hasattr(instance, 'items'):
        return {k: to_dict(v, options=options, **to_deep(include, exclude, k)) for k, v in instance.items()}

    # Any List
    if isinstance(instance, list) or hasattr(instance, '__iter__'):
        return [to_dict(x, options=options, include=include, exclude=exclude) for x in instance]

    # Additional classes:
    #  - decimal.Decimal: created by sqlalchemy.automap/reflect
    if isinstance(instance, __clsztypes__):
        return str(instance)

    # Include Columns given
    if isinstance(include, collections.Iterable):
        rtn = {}
        for column in include:
            rtn[column] = to_dict(getattr(instance, column), **to_deep(include, exclude, column))
        return rtn

    # Include all columns if it is a SQLAlchemy instance
    try:
        columns = ModelWrapper.get_columns(object_mapper(instance)).keys()
        relations = ModelWrapper.get_relations(object_mapper(instance)).keys()
        attributes = ModelWrapper.get_attributes(object_mapper(instance)).keys()
        proxies = [p.key for p in ModelWrapper.get_proxies(object_mapper(instance))]
        hybrids = [p.key for p in ModelWrapper.get_hybrids(object_mapper(instance))]
        attributes = itertools.chain(columns, relations, proxies, hybrids, attributes)
    except UnmappedInstanceError:
        raise DictConvertionError("Could not convert argument to plain dict")

    rtn = {}

    # Include AssociationProxy and Hybrids (may be list/dict/col)
    for column in attributes:

        if exclude is not None and column in exclude:
            continue
        if column in rtn:
            continue

        # Prevent unnec. db calls
        if include is False and column not in hybrids and column not in columns:
            continue

        if column not in instance.__dict__ and not options.get('execute_queries', True):
            if column not in hybrids or not options.get('execute_hybrids', True):
                continue

        # Get Attribute
        node = getattr(instance, column)

        # Don't execute queries if stopping deepnes
        if include is False and isinstance(node, Query):
            continue
        # Otherwise query it
        elif isinstance(node, Query) and options['execute_queries']:
            node = node.all()

        # Convert it
        rtn[column] = to_dict(node, **to_deep(include, exclude, column))
    return rtn



#!/usr/bin/python
# -*- encoding: utf-8 -*-
"""

"""
__author__ = 'Martin Martimeo <martin@martimeo.de>'
__date__ = '29.04.13 - 15:34'

from setuptools import setup
from tornado_restless import __version__

setup(
    name='Tornado-Restless',
    version=__version__,
    author='Martin Martimeo',
    author_email='martin@martimeo.de',
    url='https://github.com/tornado-utils/tornado-restless',
    packages=['tornado_restless'],
    license='GNU AGPLv3+ or BSD-3-clause',
    platforms='any',
    description='flask-restless adopted for tornado',
    long_description=open('README.md').read(),
    install_requires=open('requirements.txt').readlines(),
    test_suite='nose.collector',
    tests_require=open('requirements-test.txt').readlines(),
    download_url='https://github.com/tornado-utils/tornado-restless/tarball/0.4.4',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



#!/usr/bin/python
# -*- encoding: utf-8 -*-
"""

"""
from tornado.web import Application, URLSpec

from .handler import BaseHandler
from .errors import IllegalArgumentError

__author__ = 'Martin Martimeo <martin@martimeo.de>'
__date__ = '26.04.13 - 22:25'


class ApiManager(object):
    """
        The tornado restless engine

        You normally only need one instance of this class to spawn your tornado routes
    """

    METHODS_READ = frozenset(['GET'])
    METHODS_MODIFY = frozenset(['POST', 'PUT', 'PATCH'])
    METHODS_DELETE = frozenset(['DELETE'])

    METHODS_UPDATE = METHODS_READ | METHODS_MODIFY
    METHODS_ALL = METHODS_READ | METHODS_MODIFY | METHODS_DELETE

    def __init__(self,
                 application: Application,
                 session_maker: type=None):
        """
        Create an instance of the tornado restless engine

        :param session_maker: is a sqlalchemy.orm.Session class factory
        :param application: is the tornado.web.Application object
        """
        self.application = application

        self.session_maker = session_maker

    def create_api_blueprint(self,
                             model,
                             methods: set=METHODS_READ,
                             preprocessor: dict=None,
                             postprocessor: dict=None,
                             url_prefix: str='/api',
                             collection_name: str=None,
                             allow_patch_many: bool=False,
                             allow_method_override: bool=False,
                             validation_exceptions=None,
                             exclude_queries: bool=False,
                             exclude_hybrids: bool=False,
                             include_columns: list=None,
                             exclude_columns: list=None,
                             results_per_page: int=10,
                             max_results_per_page: int=100,
                             blueprint_prefix: str='',
                             handler_class: type=BaseHandler) -> URLSpec:
        """
        Create a tornado route for a sqlalchemy model

        :param model: The sqlalchemy model
        :param methods: Allowed methods for this model
        :param url_prefix: The url prefix of the application
        :param collection_name:
        :param allow_patch_many: Allow PATCH with multiple datasets
        :param allow_method_override: Support X-HTTP-Method-Override Header
        :param validation_exceptions:
        :param exclude_queries: Don't execude dynamic queries (like from associations or lazy relations)
        :param exclude_hybrids: When exclude_queries is True and exclude_hybrids is False, hybrids are still included.
        :param include_columns: Whitelist of columns to be included
        :param exclude_columns: Blacklist of columns to be excluded
        :param results_per_page: The default value of how many results are returned per request
        :param max_results_per_page: The hard upper limit of resutest per page
        :param blueprint_prefix: The Prefix that will be used to unique collection_name for named_handlers
        :param preprocessor: A dictionary of list of preprocessors that get called
        :param postprocessor: A dictionary of list of postprocessor that get called
        :param handler_class: The Handler Class that will be used in the route
        :type handler_class: tornado_restless.handler.BaseHandler or a subclass
        :return: :class:`tornado.web.URLSpec`
        :raise: IllegalArgumentError
        """
        if exclude_columns is not None and include_columns is not None:
            raise IllegalArgumentError('Cannot simultaneously specify both include columns and exclude columns.')

        table_name = collection_name if collection_name is not None else model.__tablename__

        kwargs = {'model': model,
                  'manager': self,
                  'methods': methods,
                  'preprocessor': preprocessor or {},
                  'postprocessor': postprocessor or {},
                  'allow_patch_many': allow_patch_many,
                  'allow_method_override': allow_method_override,
                  'validation_exceptions': validation_exceptions,
                  'include_columns': include_columns,
                  'exclude_columns': exclude_columns,
                  'exclude_queries': exclude_queries,
                  'exclude_hybrids': exclude_hybrids,
                  'results_per_page': results_per_page,
                  'max_results_per_page': max_results_per_page}

        blueprint = URLSpec(
            "%s/%s(?:/(.+))?[/]?" % (url_prefix, table_name),
            handler_class,
            kwargs,
            '%s%s' % (blueprint_prefix, table_name))
        return blueprint

    def create_api(self,
                   model,
                   virtualhost=r".*$", *args, **kwargs):
        """
        Creates and registers a route for the model in your tornado application

        The positional and keyword arguments are passed directly to the create_api_blueprint method

        :param model:
        :param virtualhost: bindhost for binding, .*$ in default
        """
        blueprint = self.create_api_blueprint(model, *args, **kwargs)

        for vhost, handlers in self.application.handlers:
            if vhost == virtualhost:
                handlers.append(blueprint)
                break
        else:
            self.application.add_handlers(virtualhost, [blueprint])

        self.application.named_handlers[blueprint.name] = blueprint


#!/usr/bin/python
# -*- encoding: utf-8 -*-
"""

"""
__author__ = 'Martin Martimeo <martin@martimeo.de>'
__date__ = '22.08.13 - 17:44'

from tornado.web import HTTPError


class IllegalArgumentError(ValueError, HTTPError):
    """
        An exception occuring when there was a problem in parsing arguments or model data
    """

    def __init__(self, log_message, status_code=400, *args, **kwargs):
        ValueError.__init__(self, log_message)
        HTTPError.__init__(self, status_code, log_message, *args, **kwargs)


class ProcessingException(HTTPError):
    """
        Raised when a pre or postprocessors encouters a problem.

        No more processing will be done and the processing of the query is halted
    """

    def __init__(self, log_message, status_code=400, *args, **kwargs):
        HTTPError.__init__(self, status_code, log_message, *args, **kwargs)


class DictConvertionError(HTTPError):
    """
        Raised from convert.to_dict when it can't convert an instance to plain dict
    """

    def __init__(self, instance_type, log_message=None, status_code=400, *args, **kwargs):
        super().__init__(status_code, log_message, *args, **kwargs)
        self.instance_type = instance_type


try:
    from tornado.web import MethodNotAllowedError
except ImportError:
    class MethodNotAllowedError(HTTPError):
        """An exception occuring when the method is not supported by the handler

        Takes in addition to `HTTPError` arguments method and stores it value

        :arg string method: The name of the method that was called
        """

        def __init__(self, method=None, log_message=None, status_code=405, *args, **kwargs):
            super().__init__(status_code, log_message, *args, **kwargs)
            self.method = method